use strict;
use ContractUtil;

package ContractChecks;
use Data::Dumper;

our @bhcodes = qw/BCNHGT BCNHGT2 BCNHGT3 BCNHGT4 BCNHGT5 BCNHGT6/;

sub CheckDuplicateVectorObs
{
    my($context,$report) = @_;
    return if ! $context->has('vecc_obs');
    print "Checking for duplicated vectors\n" if $context->verbose;
    $report->start_errors("Checking for duplicated vectors",
      "Checks for duplicated vectors, that is with the same start and endpoints, date/time, and observed values" );
    my %obs = ();
    foreach my $v (@{$context->vecc_obs})
    {
        my $vechash = join(':',$v->from,$v->to,@{$v->value},$v->date);
        push( @{$obs{$vechash}},$v->lineno.':'.$v->filename );
    }
    foreach my $vechash (keys %obs)
    {
        my $locations = $obs{$vechash};
        next if scalar(@$locations) == 1;
        my @data = split(/\:/,$vechash);
        $report->write_error("Vector $data[0]-$data[1] duplicated\n");
        foreach my $loc( @$locations )
        {
            @data = split(':',$loc,2);
            $report->write_error_detail("File $data[1]: line $data[0]");
        }
    }
    my $errorcount = $report->end_errors();
    print "   $errorcount duplicate vector errors\n" if $errorcount && $context->verbose;
    $report->start_summary();
    if( $errorcount )
    { 
        $report->write_error($errorcount," duplicate vector errors");
    }
    else
    {
        $report->write_info("No duplicate vector errors");
    }
    $report->end_summary();

}

sub CompareVectorObs
{
    my($context,$report) = @_;
    return if ! $context->has('vecc_obs');
    my $tolerance = $context->cfg->get('VECTOR_MATCH_TOLERANCE',0.0001);

    print "Comparing snap and contract file vector observations\n" if $context->verbose;
    $report->start_errors("Comparing SNAP and contract file vector observations",
        "Checks that the GPS observations in the supplied SNAP file match those in the CSV files. ".
        "Vectors are matched on the start and end marks, and the observed value (with a tolerance of $tolerance).  ".
        "Note that the error of the observations is not checked." );

    $tolerance *= $tolerance;

    if( ! $context->has('snap_obs') )
    {
        print "SNAP observations not available for comparison"
            if $context->verbose;
        $report->end_errors;
        $report->start_summary;
        $report->write_error("SNAP observations not available for comparison");
        $report->end_summary;
        return;
    }

    my $vectors = {};

    foreach my $obs ( @{$context->snap_obs} )
    {
        my $line = uc($obs->from.'-'.$obs->to);
        push( @{$vectors->{$line}->{snap}},$obs )
            if $obs->type eq 'GPS';
    }
    my $nobs = 0;
    foreach my $obs ( @{$context->vecc_obs} )
    {
        $nobs++;
        my $line = uc($obs->from.'-'.$obs->to);
        push( @{$vectors->{$line}->{vecc}},$obs );
    }
    my $errorcount = 0;
    foreach my $line (sort keys %$vectors )
    {
        $errorcount += 
           _CompareVectors($report,$line,$vectors->{$line}->{snap},
               $vectors->{$line}->{vecc},$tolerance);

    }
    print "   $errorcount errors matching vectors\n" if $errorcount && $context->verbose;

    $report->end_errors;
    $report->start_summary;
    if( $errorcount )
    {
        $report->write_error($errorcount," errors matching $nobs vectors");
    }
    else
    {
        $report->write_info("No errors matching $nobs vectors");
    }
    $report->end_summary;

}

sub _CompareVectors
{
    my($report,$line,$snap,$vecc,$tolerance) = @_;
    my %used = {};
    my @unmatchv = ();
    my @unmatchs = ();
    my @match = ();
    my @rejmatch = ();

    my $errorcount = 0;
    foreach my $v (@$vecc)
    {
        my $vobs = $v->value;
        my $mindiff = $tolerance * 2;
        my $nmin = -1;
        my $mindiffrej = $tolerance * 2;
        my $nminrej = -1;
        foreach my $i (0 .. $#$snap)
        {
            next if $used{$i};
            my $s = $snap->[$i];
            my $sobs = $s->value;
            my $diff = _VecDiff2($vobs,$sobs);
            next if $diff > $tolerance;
            if( $s->rejected )
            {
                if( $diff < $mindiffrej ) 
                {
                    $nminrej = $i;
                    $mindiffrej = $diff;
                }
            }
            else
            {
                if( $diff < $mindiff ) 
                {
                    $nmin = $i;
                    $mindiff = $diff;
                }
            }
        }
        $nmin = $nminrej if $nmin < 0;
        if( $nmin < 0 )
        {
            push(@unmatchv,$v);
            next;
        }
        $used{$nmin} = 1;
        my $s = $snap->[$nmin];
        my $match = [$v,$s];
        if( $s->rejected ){ push(@rejmatch,$match);}
        else { push(@match,$match); }
    }
    foreach my $i (0..$#$snap)
    {
        push(@unmatchs,$snap->[$i]) if ! $used{$i};
    }

    return $errorcount if ! @unmatchv && ! @unmatchs && ! @rejmatch;

    $report->write_error("Mismatched observations on line $line");
    if( @unmatchv )
    {
        $report->write_error_detail("Observation in vector data file is not in SNAP data file");
        foreach my $v (@unmatchv) 
        { 
            $errorcount++;
            $report->write_error_detail(_ObsDescription($v)); 
        }
    }
    if( @unmatchs )
    {
        $report->write_error_detail("Observation in SNAP data file is not listed in vector data file");
        foreach my $v (@unmatchs) 
        { 
            $errorcount++;
            $report->write_error_detail(_ObsDescription($v)); 
        }
    }
    if( @rejmatch )
    {
        $report->write_error_detail("Observation listed in vector data file is rejected in SNAP file");
        foreach my $m (@rejmatch)
        {
            $errorcount++;
            my($vobs,$sobs) = @$m;
            $report->write_error_detail("Vector obs: ",_ObsDescription($vobs));
            $report->write_error_detail("SNAP obs:   ",_ObsDescription($sobs)," [rejected]");
        }
    }
    return $errorcount;
}

sub _ObsDescription
{
    my($obs) = @_;
    my $file = $obs->filename;
    $file =~ s/.*[\\\/]//;
    my $value = $obs->value;
    return $file.":".$obs->lineno." [".join(", ",@$value)."]";
}

sub _VecDiff2
{
    my($v1,$v2) = @_;
    return ($v1->[0]-$v2->[0])**2 + ($v1->[1]-$v2->[1])**2 + ($v1->[2]-$v2->[2])**2;
}

sub CheckIndependentSessions
{
    my($context,$report) = @_;
    return if ! $context->has('vecc_obs');

    my $minduration = $context->cfg->get('INDEPENDENT_SESSION_TEST_DURATION_MINS',15);
    print "Checking for independent sessions (based on observation times)\n" if $context->verbose;
    $report->start_errors("Checking for independent sessions (based on observation times)",
        "Checks for independent sessions based on the time of observations at each mark. ".
        "Assumes that there are independent sessions if the time between the first and last observations is greater than $minduration minutes"
        );

    my $vobslist = $context->vecc_obs;
    my $range = {};
    foreach my $obs ( @$vobslist )
    {
        _AddObsDate( $range, $obs->from, $obs->date );
        _AddObsDate( $range, $obs->to, $obs->date );
    }
    my $errorcount = 0;
    my $stationcount = 0;
    foreach my $code (sort keys %$range )
    {
        $stationcount++;
        my $duration = int(($range->{$code}->{max} - $range->{$code}->{min}) / 60);
        if( $duration < $minduration )
        {
            my $count = $range->{$code}->{count};
            $errorcount++;
            $report->write_error("Station {code:$code}: Observation times only span $duration minutes ($count observations)");
        }
    }
    $report->end_errors;
    $report->start_summary;
    if( $errorcount )
    {
        my $dont = $errorcount > 1 ? "don't" : "doesn't";
        $report->write_error("$stationcount stations tested: $errorcount $dont have independent sessions");
    }
    else
    {
        $report->write_info("$stationcount stations tested: All have independent sessions");
    }
    print "   $errorcount stations don't have independent sessions\n" if $context->verbose && $errorcount;
    $report->end_summary;
}

sub CheckIndependentBaselines
{
    my($context,$report) = @_;
    return if ! $context->has('vecc_obs');

    print "Checking for independent baselines\n" if $context->verbose;
    $report->start_errors("Checking for independent baselines",
        "Checks for independent baselines at each mark. ".
        "Requires each mark to be connected to two other marks"
        );

    my $vobslist = $context->vecc_obs;
    my $connections = {};
    foreach my $obs ( @$vobslist )
    {
        $connections->{$obs->from}->{$obs->to}=1;
        $connections->{$obs->to}->{$obs->from}=1;
    }
    my $errorcount = 0;
    my $stationcount = 0;
    foreach my $code (sort keys %$connections )
    {
        $stationcount++;
        my $nbaselines=scalar(keys %{$connections->{$code}});
        if( $nbaselines < 2 )
        {
            $errorcount++;
            $report->write_error("Station {code:$code}: Only connected to one other mark");
        }
    }
    $report->end_errors;
    $report->start_summary;
    if( $errorcount )
    {
        my $plural = $errorcount > 1 ? "s" : "";
        $report->write_error("$stationcount stations tested: $errorcount station$plural are only connected to one other station");
    }
    else
    {
        $report->write_info("$stationcount stations tested: All have independent sessions");
    }
    print "   $errorcount stations are only connected to one other station\n" if $context->verbose && $errorcount;
    $report->end_summary;
}

sub _AddObsDate
{
    my($range,$code,$date) = @_;
    $range->{$code} = { min=>$date, max=>$date, count=>0 } if ! exists $range->{$code};
    $range->{$code}->{min} = $date if $date < $range->{$code}->{min};
    $range->{$code}->{max} = $date if $date > $range->{$code}->{max};
    $range->{$code}->{count}++;
}

sub CheckMarkData
{
    my($context,$report, $connected) = @_;
    return if ! $context->has('mdfc_marks');

    my $marks = $context->mdfc_marks;
    my %veccmark = ();
    if($context->has('vecc_obs'))
    {
        foreach my $obs ( @{$context->vecc_obs} )
        {
            $veccmark{uc($obs->from)} = 1;
            $veccmark{uc($obs->to)} = 1;
        }
    }
    my $mwrc_marks = $context->has('mwrc_marks') ? $context->mwrc_marks : new MarkList();
    my $relbdata=$context->has('reliability_data') ? $context->reliability_data : undef;
    my $files = $context->files;
    my %corscodes = map {$_=>1} $context->cfg->codes('cors_marks');
    my %norelbords = map {$_=>1} $context->cfg->codes('no_reliability_check_orders');

    print "Checking mark data\n" if $context->verbose;
    my $check = 
        "Checking mark data";
    my $check2 = "";
    $check2 .=
        "Checking marks against the geodetic database. ".
        "Compares the mark names and existence status. "
        if $connected;
    $check2 .= 
        "Checks that maintenance information is provided for each mark. ".
        # "Checks that reliability diagrams are provided for marks that need them.".
        "";
    $report->start_errors($check,$check2);
    my $nmark = 0;
    my %relbcodes = ();
    foreach my $mark ($marks->marks)
    {
        my $code = uc($mark->code);
        $nmark++;
        my @error = ();

        if( $connected )
        {
        my $gmark = $context->marks->mark($code)->gdb;
        if( $gmark->EXMK eq 'N' )
        {
            push(@error,"Mark is listed as existing, but isn't in geodetic database")
                if  $mark->exists;
        }
        else
        {
            push(@error,"Mark is listed as new, but is in geodetic database")
                if ! $mark->exists;
            push(@error,"Name \"".$mark->name."\" doesn't match geodetic database \"".$gmark->NAME."\"")
                if _CleanName($mark->name) ne _CleanName($gmark->NAME);
        }
        }

        my $mwrcmark = $mwrc_marks->mark($code );
        if( $mwrcmark )
        {
            if( _CleanName($mwrcmark->data->NAME) ne _CleanName($mark->name) )
            {
                push(@error,"Mark name '".$mark->name.
                      "' inconsistent with name '".
                      $mwrcmark->data->NAME."' from maintenance information");
            }
        }

        # All surveyed marks must have reliability data (16.8)

        if( ! $corscodes{$code} && $veccmark{$code} && $relbdata && ! exists $relbdata->{$code} )
        {
            push(@error,"No reliability data for mark in mark reliability file");
        }

        if( ! $corscodes{$code} && ! $mwrcmark )
        {
            push(@error,"No maintenance information for mark in RMWRC file");
        }

        if( @error )
        {
            $report->write_error("Errors for mark {code:",$code,"} from ",$mark->filename," line ", $mark->lineno, " (record ",$mark->recordno,")" );
            foreach my $error (@error) { 
                $report->write_error_detail($error); 
                $context->record_mark_error( $code, $error, " (from ",$mark->filename," line ",$mark->lineno,")");
                }
        }

    }

    my $errorcount = $report->end_errors;
    $report->start_summary;
    if( $errorcount )
    {
        $report->write_error("$nmark marks tested: $errorcount errors");
    }
    else
    {
        $report->write_info("$nmark marks tested: No errors");
    }
    print "   $nmark marks tested: $errorcount errors\n"
        if $context->verbose && $errorcount;
    $report->end_summary;
}

sub _CompileTraverseVectors
{
    my ( $code, $obs ) = @_;
    my $dtor=atan2(1.0,1.0)/45.0;
    my @unused=@$obs;
    my %vectors=();

    # print "Compile traverses:\n";
    # foreach my $o (@unused)
    # {
    #     print "  From ",$o->fref," to ",$o->tref,"\n";
    #  }

    my $haveswing=0;
    my %marks=($code=>{vector=>[0.0,0.0],paths=>1,count=>0,endpoint=>0,haveswing=>0});
    while( @unused )
    {
        my $found=0;
        foreach my $swing (0,1)
        {
            foreach my $i (0..$#unused)
            {
                my $o = $unused[$i];
                next if ! $swing && $o->bsng;
                next if $swing && ! $o->bsng;
                my $angle;
                my $from;
                my $to;
                if( exists($marks{$o->fref}) )
                {
                    $from=$o->fref;
                    $to=$o->tref;
                    $angle=$o->bear;
                }
                elsif( exists($marks{$o->tref}))
                {
                    $from=$o->tref;
                    $to=$o->fref;
                    $angle=$o->bear+180.0;
                }
                else
                {
                    next;
                }
                $found=1;
                $angle*=$dtor;
                my $dist=$o->dist;
                my $m=$marks{$from};
                $m->{endpoint}=0;
                my $haveswing = $o->bsng ? 1 : $m->{haveswing};
                my $x = $m->{vector}->[0]+cos($angle)*$dist;
                my $y = $m->{vector}->[1]+sin($angle)*$dist;
                my $count=$m->{count}+1;
                if( ! exists($marks{$to}) )
                {
                   $marks{$to}={from=>$code, to=>$to, vector=>[$x,$y],paths=>1,count=>$count,endpoint=>1,haveswing=>$haveswing, csys=>{$o->csys=>1}};
                }
                else
                {
                    my $v=$marks{$to};
                    $v->{paths}++;
                    if( ($count < $v->{count}) && ($haveswing <= $v->{haveswing}) )
                    {
                        $v->{vector}=[$x,$y];
                        $v->{count}=$count;
                        $v->{csys}->{$o->csys}++;
                    }
                }
                splice(@unused,$i,1);
                last;
            }
        last if $found;
        }
    last if ! $found;
    }
    delete $marks{$code};

#    print "="x70,"\n";
#    foreach my $k (sort keys %marks)
#    {
#        my $v=$marks{$k};
#        print "$k: $v->{from}=>$v->{to} $v->{vector}->[0] $v->{vector}->[1] E:$v->{endpoint} N:$v->{nobs} C:$v->{count} S:$v->{haveswing} P:$v->{paths}\n";
#    }

    return \%marks, \@unused;
}

sub _VecLen
{
    my ($v) = @_;
    my $ds=0;
    foreach my $dx (@$v) { $ds += $dx*$dx; }
    return sqrt($ds);
}

sub _ToleranceFunction
{
    my ($def)=@_;
    my @calcs=();
    foreach my $toldef (split(' ',$def))
    {
        if( $toldef=~/^\d+\.?\d*$/ )
        {
            push(@calcs,$toldef);
        }
        elsif( $toldef=~/^(\d\.?\d*)\+(\d\.?\d*)d$/)
        {
            push(@calcs,"$1 + $2*\$d");
        }
        elsif( $toldef=~/^(\d\.?\d*)\+\-(\d\.?\d*)d$/)
        {
            my ($v1,$v2)=($1*$1,$2*$2);
            push(@calcs,"sqrt($v1 + $v2*\$d*\$d)");
        }
    }
    my $func='sub (){ my $d=$_[0]; my @v=sort('.join(',',@calcs).'); return $v[0]}';
    my $sub=eval($func);
    return eval($func);
}

sub CheckReliabilityData
{
    my($context,$report, $connected) = @_;
    return if ! $context->has('reliability_data');

    my %marknamecode=();
    if( $context->has('mdfc_marks') )
    {
        foreach my $mark (@{$context->mdfc_marks->marks})
        {
            $marknamecode{$mark->name}=$mark->code;
        }
    }

    my $relbdata=$context->reliability_data;

    my $marks=$context->marks;

    my $visokmap=$context->cfg->mapping('VISUAL_RELIABILITY_OK');
    my $visok={};
    foreach my $k (keys %$visokmap)
    {
        my $orders={};
        foreach my $o (split(' ',$visokmap->{$k})){ $orders->{$o}=1; }
        $visok->{$k}=$orders;
    }

    print "Checking mark reliability data\n" if $context->verbose;
    $report->start_errors("Checking mark reliability data",
        "Checks that the reliability check types are valid, and that ".
        "traverse observations are consistent");

    my $vtolfunc = _ToleranceFunction($context->cfg->relb_traverse_vector_tolerance);
    my $stolfunc = _ToleranceFunction($context->cfg->relb_traverse_length_tolerance);
    my $nmark=0;

    my $nrelbadpt=0;
    my $nrelbvisl=0;
    my $nrelbnewm=0;
    my $nrelbobsc=0;

    my $nfailadpt=0;
    my $nfailvisl=0;
    my $nfailnewm=0;
    my $nfailobsc=0;
    
    my $ninfo=0;
    my $nzgd2000=$context->cslist->coordsys('NZGD2000');

    foreach my $code ( keys %$relbdata )
    {
        $nmark++;
        my @errors=();
        my @info=();
        my $relb=$relbdata->{$code};
        my $markinfo = $marks->mark($code);
        my $mmark=$markinfo && $markinfo->has('mdfc') ? $markinfo->mdfc : undef;
        my $gmark=$markinfo && $markinfo->has('gdb') ? $markinfo->gdb : undef;
        my $obs=$relb->obs;
        my $types={};
        my %reported=();
        foreach my $o (@$obs)
        {
            my $type=$o->type;
            foreach my $nametype (qw/FREF TREF/)
            {
                my $nt=lc($nametype);
                my $name=$o->{$nt};
                my $mcode=$marknamecode{$name};
                next if $name eq '' || $mcode eq '';
                $o->{$nt}=$mcode;
                next if $reported{$mcode};
                $reported{$mcode}=1;
                push(@errors,"Mark name used instead of code for $mcode ($name)");
            }
            $types->{$type}=[] if ! exists $types->{$type};
            push(@{$types->{$type}},$o);
        }
        my @typecodes=sort keys %$types;

        # Check for valid combination of codes...
        my $codesvalid=scalar(@typecodes)==1 || 
            (scalar(@typecodes)==2 && $typecodes[0] eq 'ADPT' && $typecodes[1] eq 'OBSA');
        if( ! $codesvalid )
        {
            push(@errors,"Inconsistent reliability check types ".join(' ',@typecodes));
        }

        if( scalar(@typecodes)==1 && ($typecodes[0] eq 'ADPT' || $typecodes[0] eq 'OBSA'))
        {
            push(@errors,$typecodes[0]." reliability check must include both OBSA and ADPT data")
        }

        # Check whether NEWM type check is consistent with mark status
        $nrelbnewm++ if exists $types->{'NEWM'};
        if( exists $types->{'NEWM'} && $mmark && $mmark->exists )
        {
            push(@errors,"NEWM reliability type not valid for existing mark");
            $nfailnewm++;
        }

        # Check if VISL type check is consistent with coordinate orders.
        $nrelbvisl++ if exists $types->{'VISL'};
        if( exists $types->{'VISL'} && $gmark )
        {
            my $ok=0;
            foreach my $k (keys %$visok)
            {
                $ok=1 if $gmark->has($k) && exists($visok->{$k}->{$gmark->{$k}});
            }
            if( ! $ok )
            {
                push(@errors,"VISL reliability type not valid for mark coordinate order");
                $nfailvisl++;
            }
        }

        # Check that traverse information is a set of traverses to the mark...

        my $traverse={};
        foreach my $type (qw/OBSC OBSA ADPT/)
        {
            next if ! exists $types->{$type};
            my ($vectors,$unused)=_CompileTraverseVectors($code,$types->{$type});
            if( ! %$vectors )
            {
                my $nunused = scalar(@$unused);
                push(@errors,"Reliability observation of type $type do not connect to $code ($nunused obs)");
            }
            else
            {
                foreach my $o (@$unused)
                {
                    push(@errors,"Reliability observation $type from ".$o->fref." to ".$o->tref." not used in a traverse");
                }
            }
            $traverse->{$type}=$vectors;
        }

        # Now how about checking the traverses actually match coordinates or each
        # other..
        
        if( exists($traverse->{OBSC}) )
        {
            my $obsc=$traverse->{OBSC};
            $report->write_info("OBSC check: No gdb info for $code to check OBSC") if ! $gmark;

            foreach my $to (keys %$obsc)
            {
                $nrelbobsc++;
                $nfailobsc++;
                my $tmarkinfo = $marks->mark($to);
                my $tmmark=$tmarkinfo && $tmarkinfo->has('mdfc') ? $markinfo->mdfc : undef;
                my $tgmark=$tmarkinfo && $tmarkinfo->has('gdb') ? $tmarkinfo->gdb : undef;
                $report->write_error("OBSC check: No gdb info for $to to check OBSC") if ! $tgmark;
                next if ! $gmark || ! $tgmark;
                my $fcrd=$nzgd2000->coord($gmark->CRDLAT,$gmark->CRDLON,$gmark->CRDHGT);
                my $tcrd=$nzgd2000->coord($tgmark->CRDLAT,$tgmark->CRDLON,$tgmark->CRDHGT);
                my @csyslist=keys %{$obsc->{$to}->{csys}};
                my $csys=$csyslist[0];
                if( scalar(@csyslist) > 1 )
                {
                    $report->write_error("Traverse data from $code to $to uses to multiple coordinate systems (".
                        join(',',@csyslist).") - using $csys");
                }
                my $cs=$context->cslist->coordsys($csys);
                if( ! $cs )
                {
                    $report->write_error("Traverse data from $code to $to uses unknown coordinate system $csys");
                    next;
                }
                my $fenh=$fcrd->as($cs);
                my $tenh=$tcrd->as($cs);
                my $vgdb=[$tenh->[0]-$fenh->[0],$tenh->[1]-$fenh->[1]];
                my ($scalefactor,$convergence) = @{$fenh->sf_conv()};
                my $vector=$obsc->{$to}->{vector};
                # Convert observed vector from ellipsoidal to projection distance.
                #$report->write_info(Dumper($vgdb,$vector,$scalefactor));
                my $vobsc=[$vector->[0]*$scalefactor,$vector->[1]*$scalefactor];
                my $dx=$vobsc->[0]-$vgdb->[0];
                my $dy=$vobsc->[1]-$vgdb->[1];
                my $vdiff=_VecLen([$dx,$dy]);
                my $sdiff=abs(_VecLen($vobsc)-_VecLen($vgdb));
                my $length=_VecLen($vobsc);
                my $vtol=$vtolfunc->($length);
                $nfailobsc--;
                next if $vdiff < $vtol;
                my $stol=$stolfunc->($length);
                next if $obsc->{$to}->{haveswing} && $sdiff < $stol;
                $nfailobsc++;
                if( $sdiff < $stol )
                {
                    push(@errors,"OBSC traverse gdb misclose ".
                        sprintf("%.3f",$vdiff).
                        " to $to doesn't meet tolerance ".
                        sprintf("%.3f",$vtol).
                        " but length difference ".
                        sprintf("%.3f",$sdiff). 
                        " is within tolerance");
                }
                else
                {
                    push(@errors,"OBSC traverse gdb misclose ".
                        sprintf("%.3f",$vdiff).
                        " to $to doesn't meet tolerance ".
                        sprintf("%.3f",$vtol));
                }

            }
        }

        if( exists($traverse->{OBSA}) || exists ($traverse->{ADPT}))
        {
            $nrelbadpt++;
            $nfailadpt++;
        }
        if( exists($traverse->{OBSA}) && exists ($traverse->{ADPT}) )
        {
            my $obsa=$traverse->{OBSA};
            my $adpt=$traverse->{ADPT};
            my $passed=0;
            my $failed=0;

            foreach my $to (keys %$obsa)
            {
                next if ! exists $adpt->{$to};
                $passed++;
                my $vobsa=$obsa->{$to}->{vector};
                my $vadpt=$adpt->{$to}->{vector};
                my $dx=$vobsa->[0]-$vadpt->[0];
                my $dy=$vobsa->[1]-$vadpt->[1];
                my $vdiff=_VecLen([$dx,$dy]);
                my $sdiff=abs(_VecLen($vobsa)-_VecLen($vadpt));
                my $length=_VecLen($vobsa);
                my $vtol=$vtolfunc->($length);
                next if $vdiff < $vtol;
                my $stol=$stolfunc->($length);
                next if $obsa->{$to}->{bsng} && $sdiff < $stol;
                $failed++;
                if( $sdiff < $stol )
                {
                    push(@errors,"Traverse closure ".
                        sprintf("%.3f",$vdiff).
                        " to $to doesn't meet tolerance ".
                        sprintf("%.3f",$vtol).
                        " but length difference ".
                        sprintf("%.3f",$sdiff). 
                        " is within tolerance");
                }
                else
                {
                    push(@errors,"Traverse closure ".
                        sprintf("%.3f",$vdiff).
                        " to $to doesn't meet tolerance ".
                        sprintf("%.3f",$vtol));
                }
            }
            foreach my $to (keys %$obsa)
            {
                next if ! $obsa->{$to}->{endpoint};
                next if exists $adpt->{$to};
                push(@errors,"No adopted traverse data to observed mark $to");
            }
            foreach my $to (keys %$adpt)
            {
                next if ! $adpt->{$to}->{endpoint};
                next if exists $obsa->{$to};
                push(@errors,"No observed traverse data to adopted mark $to");
            }
            foreach my $to (keys %$obsa)
            {
                next if $obsa->{$to}->{paths} < 2;
                $ninfo++;
                $report->write_info("More than one observation to $to");
            }
            foreach my $to (keys %$adpt)
            {
                next if $adpt->{$to}->{paths} < 2;
                $ninfo++;
                $report->write_info("More than one adoption to $to");
            }
            $nfailadpt-- if $passed && ! $failed;
        }

        if( @errors )
        {
            $report->write_error("Errors in reliability data for $code");
            foreach my $error (@errors) 
            { 
                $report->write_error_detail($error); 
                $context->record_mark_error( $code, $error, " (reliability data)");
            }
        }
    }

    my $errorcount = $report->end_errors;
    $report->start_summary;
    if( $errorcount )
    {
        $report->write_error("Data for $nmark marks tested: $errorcount errors");
    }
    elsif( $ninfo )
    {
        $report->write_error("Data for $nmark marks tested: $ninfo warnings");
    }
    else
    {
        $report->write_info("Data for $nmark marks tested: No errors");
    }
    if( $nrelbnewm )
    {
        if( $nfailnewm )
        {
            $report->write_error("Failed $nfailnewm of $nrelbnewm NEWM reliability checks");
        }
        else
        {
            $report->write_error("Passed $nrelbnewm NEWM reliability checks");
        }
    }
    if( $nrelbvisl )
    {
        if( $nfailvisl )
        {
            $report->write_error("Failed $nfailvisl of $nrelbvisl VISL reliability checks");
        }
        else
        {
            $report->write_error("Passed $nrelbvisl VISL reliability checks");
        }
    }
    if( $nrelbobsc )
    {
        if( $nfailobsc )
        {
            $report->write_error("Failed $nfailobsc of $nrelbobsc OBSC reliability checks");
        }
        else
        {
            $report->write_error("Passed $nrelbobsc OBSC reliability checks");
        }
    }
    if( $nrelbadpt )
    {
        if( $nfailadpt )
        {
            $report->write_error("Failed $nfailadpt of $nrelbadpt OBSA/ADPT reliability checks");
        }
        else
        {
            $report->write_error("Passed $nrelbadpt OBSA/ADPT reliability checks");
        }
    }

    print "   $nmark marks tested: $errorcount errors\n"
        if $context->verbose && $errorcount;
    $report->end_summary;

}

sub GetSummaryReportSheets
{
    my( $context, $report ) = @_;
    my $files = $context->files->find('','MSRX1');
    if( ! $files )
    {
        $report->write_error("There is no maintenance summary report (excel file)");
        return;
    }
    my $codes=[];
    my $book;
    eval
    {
        my $excel;
        eval
        {
            require Win32::OLE;
            $Win32::OLE::Warn = 3; # die on errors
        };
        if( $@ )
        {
            die "Cannot read Excel file: Win32::OLE module not available\n";
        }
        eval
        {
            $excel = Win32::OLE->new('Excel.Application', sub { $_[0]->Quit; } );
        };
        if( $@ )
        {
            die "Cannot read Excel file: Excel.Application OLE class not defined\n";
        }
        foreach my $f (@$files)
        {
            my $book = $excel->Workbooks->Open($f->path);
            foreach my $i (1 .. $book->Worksheets->Count)
            {
                push(@$codes,lc($book->Worksheets($i)->Name));
            }
            $book->{Saved}=1;
            undef $book;
        }
        $excel->Quit();
        undef $excel;
    };
    if( $@ )
    {
        $report->write_error("Cannot read excel maintenance summary report: $@");
        $codes = undef;
    }
    $book->Close if $book;
    return if ! $codes;
    return wantarray ? @$codes : $codes;
}

sub CheckMaintenanceData
{
    my($context,$report,$connected) = @_;
    return if ! $context->has('mwrc_marks');
    my $marks = $context->mwrc_marks;
    my $files = $context->files;
    my $bhgttol1 = $context->cfg->get('BEACON_HEIGHT_TOLERANCE',0.10);
    my $bhgttol2 = $context->cfg->get('SAME_BEACON_HEIGHT_TOLERANCE',0.01);

    print "Checking mark maintenance\n" if $context->verbose;
    my $description = 
        "Checks that beacon heights are provided and that they are self consistent - ie in the correct order of height.\n\n".
        "Checks that the heights match the geodetic database when the beacon date is not altered (tolerance $bhgttol2).\n\n".
        "Checks that required photographs and and diagrams have been provided.\n".
        "";

    if( ! $connected )
    {
        $description .= "\n\nCannot check maintenance data gdb - not connected to database\n";
        $report->write_info("Cannot check maintenance data against gdb - not connected to database");
    }
    $report->start_errors("Checking mark maintenance data",$description);

    my $reportsheets =  GetSummaryReportSheets($context,$report);
    if( $reportsheets )
    {
        my %sheets = map { lc($_)=>1 } @$reportsheets;
        $reportsheets = \%sheets;
    }
    my $reportfiles = join(' ',map {$_->name} $files->find('','MSRX1'));

    my $nmark = 0;
    my $nerror = 0;
    my $nwarn = 0;
    my %photocodes=();
    my %adcodes=();
    my %bdcodes=();
    foreach my $mark ($marks->marks)
    {
        $nmark++;
        my @errors = ();
        my @info = ();

        my $data = $mark->data;

        if( $data->MPSC eq 'NFND' )
        {
            push(@info,"Mark not found");
        }
        elsif( $data->MPSC eq 'DEST')
        {
            push(@info,"Mark recorded as destroyed");
        }
        else
        {
        if( $data->MPSC eq 'NSPE')
        {
            push(@info,"Mark condition not specified");
        }

        # Check beacon details

        my $btype = $data->MRKE;

        push(@errors,"BDAT (beacon date) missing (beacon type = ".$data->MRKE.")")
            if $data->BDAT eq '' && $btype=~/^(2M|4M|NS)$/;

        my($berrors,$binfo,$bhgts) = CheckBeaconHeights($data);
        push(@errors,@$berrors);
        push(@info,@$binfo);


        # Check for maintenance info

        if( $reportsheets && $mark->maintained && ! $reportsheets->{lc($mark->code)})
        {
            push(@errors,"Required maintenance report not found in $reportfiles");
        }

        # Check for site photo
    
        my $codere = quotemeta($mark->code);
        $photocodes{uc($mark->code)} = 1;
        my $sitephotos = $files->find($codere."\\d\\dP\\d\\.JPG");
        if( $sitephotos )
        {
            foreach my $f (@$sitephotos)
            {
                $context->add_mark_image($mark->code,"Site photo",$f->path);
            }
        }
        else
        {
            push(@errors,"Site photo not supplied\n");
            push(@{$context->missinginfo},"Site photograph for mark {code:".$mark->code."}" );
        }

        # Check for beacon diagram

        if( $btype eq 'NS' )
        {
            $bdcodes{uc($mark->code)} = 1;
            my $beacdiags = $files->find($codere."\\d\\dBD\\.PNG");
            if( $beacdiags )
            {
                foreach my $f (@$beacdiags)
                {
                    $context->add_mark_image($mark->code,"Beacon diagram",$f->path);
                }
            }
            else
            {
                push(@errors,"Beacon diagram not supplied\n");
                push(@{$context->missinginfo},"Beacon diagram for mark {code:".$mark->code."}" );
            }
        }

        # Check for access or finder diagrams
        
        $adcodes{uc($mark->code)} = 1;
        my $accdiags = $files->find($codere."\\d\\dAD\\.PNG");
        if( $accdiags )
        {
            foreach my $f (@$accdiags)
            {
                $context->add_mark_image($mark->code,"Access diagram",$f->path);
            }
        }
        else
        {
            push(@errors,"Access diagram not supplied\n");
            push(@{$context->missinginfo},"Access diagram for mark {code:".$mark->code."}" );
        }


        if( $connected )
        {
            my $exists = uc($data->EXMK) eq 'Y';

            my $gmark = $context->marks->mark($mark->code)->gdb;
            if( $gmark->EXMK eq 'N' )
            {
                push(@errors,"Mark is listed as existing, but isn't in geodetic database")
                    if  $exists;
            }
            else
            {
                push(@errors,"Mark is listed as new, but is in geodetic database")
                    if ! $exists;
                push(@errors,"Name \"".$data->NAME."\" doesn't match geodetic database \"".$gmark->NAME."\"")
                    if _CleanName($data->NAME) ne _CleanName($gmark->NAME);

                push(@info,"Mark type ".$data->MRKT." differs from geodetic database ".$gmark->MRKT)
                    if $gmark->MRKT && $gmark->MRKT ne 'OTHR' && $gmark->MRKT ne 'UNKN'
                      && $data->MRKT ne $gmark->MRKT;

                push(@info,"Mark protection type ".$data->MRKR." differs from geodetic database ".$gmark->MRKR)
                    if $gmark->MRKR && $gmark->MRKR ne 'NSPE' && $data->MRKR ne $gmark->MRKR;

                push(@info,"Mark beacon type ".$data->MRKE." differs from geodetic database ".$gmark->MRKE)
                    if $gmark->MRKE && $gmark->MRKE ne 'ND' && $data->MRKE ne $gmark->MRKE;

                push(@info,"Ground level relationship ".$data->GLREL." differs from geodetic database ".$gmark->GLREL)
                    if $gmark->GLREL ne '' && $data->GLREL != $gmark->GLREL;

                my $samebeacon = ContractUtil::CalcDate($gmark->BDAT) >= ContractUtil::CalcDate($data->BDAT);
                my $tol = $samebeacon ? $bhgttol2 : $bhgttol1;
                my $bherror = $samebeacon ? \@errors : \@info;
                my $bhmsg = $samebeacon ? "" : " significantly";
                for my $ih ( 0 .. $#$bhgts )
                {
                    my $ghgt = $gmark->{$bhcodes[$ih]};
                    push( @$bherror,"Beacon height $bhcodes[$ih] $bhgts->[$ih] differs$bhmsg from geodetic database $ghgt")
                        if $bhgts->[$ih] ne '' && $ghgt ne '' && abs($bhgts->[$ih]-$ghgt) > $tol;
                }
                push(@info,"GL relationship ".$data->GLREL." should be less than zero")
                   if $data->GLREL > 0.0 && $data->MRKE eq 'PL';
            }
            
        }
        }

        my $markdesc = join('',"mark {code:",$mark->code,"}",
            " (",$data->NAME,") state ",$data->MPSC,
            " from ",$mark->filename," line ", $mark->lineno,
            " (record ",$mark->recordno,")");

        if( @errors )
        {
            $nerror++;
            $report->write_error("Errors for $markdesc");
            foreach my $error (@errors) 
            { 
                $report->write_error_detail($error); 
                $context->record_mark_error($mark->code,$error," (from ",
                    $mark->filename," line ",$mark->lineno,")");
            }
        }
        if( @info )
        {
            $nwarn++ if ! @errors;
            $report->write_info("Warnings for $markdesc")
                if ! @errors;
            foreach my $warning (@info)
            { 
                $report->write_info_detail($warning); 
                $context->record_mark_error($mark->code,'info',
                    $warning," (from ", $mark->filename,
                    " line ",$mark->lineno,")");
            }
        }

    }

    foreach my $photo ($files->find("....\\d\\dP\\d\\.JPG"))
    {
        my $code = uc(substr($photo->name,0,4));
        $report->write_error("Unused site photo ",$photo->name)
            if ! $photocodes{$code};
    }

    foreach my $ad ($files->find("....\\d\\dAD\\.PNG"))
    {
        my $code = uc(substr($ad->name,0,4));
        $report->write_error("Unused access diagram ",$ad->name)
            if ! $adcodes{$code};
    }

    foreach my $bd ($files->find("....\\d\\dBD\\.JPG"))
    {
        my $code = uc(substr($bd->name,0,4));
        $report->write_error("Unused non-standard beacon diagram ",$bd->name)
            if ! $bdcodes{$code};
    }

    my $errorcount = $report->end_errors;
    $report->start_summary;
    if( $nerror )
    {
        $report->write_error("$nmark marks tested: $nerror have errors, $nwarn have warnings");
    }
    elsif( $nwarn )
    {
        $report->write_info("$nmark marks tested: $nwarn have warnings");
    }
    else
    {
        $report->write_info("$nmark marks tested: No errors");
    }
    print "   $nmark marks tested: $nerror have errors, $nwarn have warnings\n"
        if $context->verbose && ($nerror || $nwarn);
    $report->end_summary;
}


sub CheckBeaconHeights
{
    my($mark) = @_;
    my @errors=();
    my @info=();
    push(@info,"GLREL (ground height relationship) missing")
        if $mark->GLREL eq '';
    push(@errors,"BDAT (beacon date) missing (beacon type = ".$mark->MRKE.")")
        if $mark->BDAT eq '' && $mark->MRKE=~/^(2M|4M|NS)$/;
    my $btype = $mark->MRKE;
    my $nhgt = 0;
    $nhgt = 6 if $btype eq '2M' || $btype eq '4M';
    # Beacons less than 1m high don't need heights.
    $nhgt = 3 if $btype eq 'PL' && $mark->GLREL <= -1.0;
    my @bhgts = map {$mark->get($_)} @bhcodes;
    my $ok = 1;
    my $lasthgt = '';
    my @missing = ();
    my $maxhgt = 0.0;
    for my $ih (0..$nhgt-1)
    {
        if( $bhgts[$ih] eq '' )
        {
            push(@missing,$bhcodes[$ih]);
            next;
        }
        $ok = 0 if $lasthgt ne '' && $bhgts[$ih] > $lasthgt;
        $maxhgt = $bhgts[$ih] if $bhgts[$ih] > $maxhgt;
        $lasthgt = $bhgts[$ih];
    }
    @bhgts = @bhgts[0 .. $nhgt-1];
    push(@errors,"Pillar heights cannot be positive\n") if $btype eq 'PL' and $maxhgt > 0.0;
    push(@info,"Missing beacon height ".join(' ',@missing)) if @missing;
    push(@errors,"Beacon heights (".join(", ",@bhgts).") are not in correct order\n") if ! $ok;
    return \@errors, \@info, \@bhgts;
}
   
sub _CleanName
{
    my($name) = @_;
    $name = uc($name);
    $name =~ s/^\s+//;
    $name =~ s/\s+$//;
    $name =~ s/\s+/ /;
    return $name;
}

1;

