
#===============================================================================
#
#   Class:       ContractConfig;
#
#   Extension of LINZ::Config
#
#   Description: Defines the following routines:
#                  $cfg = new ContractConfig()
#                  $value = $cfg->item
#                  $value = $cfg->get($item, $default)
#                  # Return list of codes or keys of mapping
#                  @codes = $cfg->codes($codetype)
#                  # Regexp string matching any of specifed codes
#                  $cfg->codeRegexp($codetype)
#                  $mapping = $cfg->mapping($maptype);
#                  $value = $mapping->{$key};
#                  $value = $cfg->map($maptype,$value);
#                  $revmap = $cfg->mapping($maptype,1);
#                  $key = $cfg->unmap($maptype,$key);
#                  $acc = $cfg->accuracy($class);
#                  $lineacc = $acc->{mm}+$length*$acc->{ppm}/1000.0;
#                  $dir = $cfg->image_directory($code);
#
#  Configuration may include
#  SOMETHING_CODES <<END_CODES
#    CODE1
#    CODE2
#  END_CODES
#
#  SOMETHING_MAP <<END_MAP
#    KEY1 VALUE1
#    KEY2 VALUE2
#  END_MAP
#
#  ACCURACY_MAP <<END_MAP
#     class1 mm_value1 ppm_value1
#     class2 mm_value2 ppm_value2
#  END_MAP
#
#===============================================================================

use strict;
use ContractUtil;

package ContractConfig::Accuracy;
use base 'AutoHash';

package ContractConfig;
use base 'LINZ::Config';

our $_cfg;

sub new 
{
    my ($class,$cfgext) = @_;
    my $cfgpath=$ENV{GEODETIC_CONTRACT_CONFIG} || 'ContractConfig.cfg';
    if($cfgext || ! $_cfg)
    {
        $_cfg = LINZ::Config::new(
            $class,
            _configpath=>$cfgpath,
            _configextra=>$cfgext,
            _case_sensitive=>0,
            _codes=>{},
            _mappings=>{},
            _revmappings=>{},
            _accuracy=>{},
          );
        print "*** USING TEST CONFIGURATION ***\n"
           if $_cfg->get('test_configuration',0);
   }
   return $_cfg;
}


sub codes
{
    my($self,$type) = @_;
    $type = uc($type);
    if( ! exists $self->{_codes}->{$type} )
    {
        my @codes=();
        my $codestr = $self->get($type.'_CODES','');
        if( $codestr )
        {
            @codes = split(' ',$self->get($type.'_CODES',''));
        }
        else
        {
            my $map = $self->mapping($type);
            @codes = sort keys %$map;
        }
        $self->{_codes}->{$type} = \@codes;
    }
    my $codes = $self->{_codes}->{$type};
    return wantarray ? @{$codes} : $codes;
}

sub codeRegexp {
    my($self,$group,$optional) = @_;
    my $opt = $optional ? '?' : '';
    return '^(?:'.join('|',map {quotemeta($_)} $self->codes($group)).')'.$opt.'$';
    }

sub mapping
{
    my ($self,$type,$reverse) = @_;
    $type = uc($type);
    if( ! exists $self->{_mappings}->{$type} )
    {
        my $mapping = {};
        my $mapdef = $self->get($type.'_MAP','');
        foreach my $l (split(/\n/,$mapdef))
        {
            #next if $l =~ /^\s*(\#|$)/;
            my ($key,$value) = split(' ',$l,2);
            next if $key !~ /\w+/;
            next if $value !~ /\S/;
            $value=$1 if $value =~ /^\'(.*)\'$/;
            $value =~ s/\s*$//;
            $mapping->{$key} = $value;

        }
        $self->{_mappings}->{$type} = $mapping;
    }
    return $self->{_mappings}->{$type} if ! $reverse;
    if( ! exists $self->{_revmappings}->{$type} )
    {
        my $mapping = $self->{_mappings}->{$type};
        my $revmapping={};
        foreach my $key (sort keys %$mapping)
        {
            my $value = $mapping->{$key};
            $revmapping->{$value} = $key if ! exists $revmapping->{$value};
        }
        $self->{_revmappings}->{$type} = $revmapping;
    }
    return $self->{_revmappings}->{$type};
}

sub map
{
    my($self,$type,$key) = @_;
    return $self->mapping($type)->{$key};
}

sub unmap
{
    my($self,$type,$key) = @_;
    return $self->mapping($type,1)->{$key};
}

# Application specific utilities

# Interprets accuracy mapping as mm and ppm errors

sub accuracy
{
    my($self,$class) = @_;
    $class = uc($class);
    return $self->{_accuracy}->{$class} if exists $self->{_accuracy}->{$class};
    my $accdef = $self->map('accuracy',$class);
    my $accuracy = undef;
    $accuracy = new ContractConfig::Accuracy(mm=>$1,ppm=>$2)
        if $accdef =~ /^([\d\.]+)\s+([\d\.]+)\s*$/i;
    $self->{_accuracy}->{$class} = $accuracy;
    return $accuracy;
}

# Determines image directory for a given geodetic code

sub image_directory
{
    my($self,$code) = @_;
    $code = uc($code);
    my $dir = $self->image_home;
    $dir .= "/marks";
    for my $i (1..4)
    {
        $dir .= "/".substr($code,0,$i);
    }
    return $dir;
}

1;
