use strict;
use ContractUtil;
use ContractFiles;

package FileCheck;
use base 'AutoHash';

sub new 
{
    my($class,$context, $type ) = @_;
    my $cfg = $context->cfg;
    my $filetype = ContractFileType::filetype($type);
    return AutoHash::new($class,
        context=>$context,
        cfg=>$cfg,
        filetype=>$filetype
    )
}

sub check
{
    my($self,$file,$report) = @_;
    $self->set_report($report);
    $self->set_file($file);

    my $name = $file->name;
    my $typename = $self->filetype->name;

    print "Checking $typename $name\n" if $self->context->verbose;
    $self->report->start_errors("Checking $typename $name");
    if( ! $self->filetype->nameIsValid($name) )
    {
        $self->report_error($name," is not a valid name for a ",$typename," file");
    }
    $self->checkfile($file);
    $self->checkfiledata;
    $self->report->end_errors;
    $self->report->start_summary;
    $self->summary($file);
    $self->report->end_summary;
    $self->set_file(undef);
}

sub check_files
{
    my( $self, $type, $report ) = @_;
    foreach my $file (@{$self->context->files->files})
    {
        $self->check($file,$report) 
           if $file->type && $file->type eq $type;
    }
}

# Override in derived classes
sub checkfile
{
    my ($self) = @_;
    die "checkfile not implemented for class ",ref($self),"\n";
}

# Override to apply checks to file data after file completely read
sub checkfiledata
{
}


# Override for additional summary information derived classes 

sub summary
{
    my ($self,$file) = @_;
    my $nrec = $self->{nrec};
    if( $self->report->error_count)
    {
       $self->report_error($self->report->error_count," errors in file");
       print "   ",$self->report->error_count," errors in file\n" if $self->context->verbose;

    }
    else
    {
       $self->report_info("No errors in file");
    }
}

# Functions used by derived classes for reporting

sub report 
{
    my($self) = @_;
    return $self->{report} || $self->context->report;
}

sub report_info
{
    my $self = shift(@_);
    $self->report->write_info(@_);
}

sub report_error
{
    my $self = shift(@_);
    $self->report->write_error(@_);
}

# Utility functions

sub file_type_name
{
    my( $self,$type ) = @_;
    return $self->cfg->get($type.'_TYPE_NAME',$type);
}

package CsvFileCheck;
use base 'FileCheck';
use CSV_File;

sub checkfile
{
    my($self,$file) = @_;

    eval
    {
        my $filetype = $self->filetype;
        my $csv = new CSV_File($file->path,$filetype->csvopts);
        $self->{csv} = $csv;

        my ($colerror,$missingcols) = $filetype->missingCsvColumns($csv);
        $self->report_error($colerror) if $colerror;

        $self->{nrec} = 0;
        while( my $record = $csv->NextRecord )
        {
            $self->{nrec}++;
            foreach my $error ($filetype->errorsInRecord($record,$missingcols))
            {
                $self->report_error($error," at line ",$csv->lineno,
                    " (record ",$csv->recordno,")")

            }
            eval
            {
                $self->checkrecord($csv->record);
            };
            if( $@ )
            {
                $self->report_error("Failed checking ",$file->name,
                    " (record ",$csv->recordno,"): ",$@);
            }
        }
    };
    if( $@ )
    {
        $self->report_error("Failed checking ",$file->name,": ",$@);
    }

}
sub checkrecord
{
    # my($self,$data) = @_;
    return;
}

sub summary
{
    my($self,$file) = @_;
    $self->report_info($self->{nrec}," records in file");
    $self->SUPER::summary($file);
}

package IndexFileCheck;
use base 'CsvFileCheck';

sub new
{
    my ( $class, $context ) = @_;
    my $self = bless CsvFileCheck->new($context,'IDX'), $class;
    $self->{filetypesfound} = {};
    return $self;
}

sub check_file_name
{
    my($self,$name, $type) = @_;
    my $rtype = $type;
    my $ftype = ContractFileType::filetype($type);
    my $error = '';
    my $info = 0;
    $name =~ s/.*[\\\/]//;
    if( ! $ftype->defined )
    {
        $error = "File type $type of $name is not recognized\n";
        my $itype = ContractFileType::inferredType( $name );
        $error .= ' - assumed to be '.$itype->type if $itype;
        $rtype = $itype->type if $itype;
    }
    elsif( ! $ftype->nameIsValid($name))
    {
        $error = "$name is not a valid filename for a ".
            $ftype->type." (".$ftype->name.") file";
        $info = 1;
    }
    if( $error )
    {
        $error .= '.';
        if( $info )
        {
            $self->report_info($error);
        }
        else
        {
            $self->report_error($error);
        }
    }
    $self->{filetypesfound}->{$rtype} = 1;
    return $rtype;
}

sub checkfiledata
{
    my($self) = @_;
    foreach my $file ( @{$self->context->files->files} )
    {
        next if $file->type;
        next if $file->name eq $self->file->name;
        my $ft = ContractFileType::inferredType($file->name);

        next if $ft && $ft->type eq 'IGNORE';
        $self->report_info( "Supplied file ", $file->name," is not defined in index file\n");
    }
    my $jobtype = $self->context->jobtype;
    if( $jobtype )
    {
        my $types = {};
        for my $t (keys %{$self->{filetypesfound}})
        {
            $types->{$t} = 1;
            $t =~ s/\d$/#/;
            $types->{$t} = 1;
        }
        foreach my $rt (@{$jobtype->reqfiles})
        {
            my @rtlist=split(/\|/,$rt);
            my @found = grep { $types->{$_}} @rtlist;
            next if @found;
            @rtlist=map {s/\#$//; $_} @rtlist;
            @rtlist=map {ContractFileType::filetype($_)->name} @rtlist;
            my $names=join(' or ',@rtlist);
            $self->report_error("Required $names not supplied (or not correctly named)")
        }
    }
}

sub checkrecord
{
    my($self,$data) = @_;
    my $daft = $data->DAFT;
    my $name = $data->DAFL;
    $daft = $self->check_file_name($name,$daft);

    my $files=$self->context->files;
    my $file = $files->file($name);
    if( ! $file )
    {
        $self->report_error("File ",$name," listed in index file has not been supplied.");
        push( @{$self->context->missinginfo},"File ".$name." listed in index file");
        return;
    }
    if( $file->type )
    {
        $self->report_error("File ",$name," duplicated in index file");
    }
    $file->{type} = $daft;
    $file->{indexdata} = $data;
}

package VeccObs;
use base 'AutoHash';

package VeccFileCheck;
use base 'CsvFileCheck';

sub new
{
    my ($class, $context ) = @_;
    return bless CsvFileCheck->new($context,'VECC1'), $class;
}

sub checkrecord
{
    my($self,$data) = @_;
    my $csv = $self->{csv};
    my $obs = new VeccObs(
        from => $data->FCODE,
        to => $data->TCODE,
        date => ContractUtil::CalcDate($data->DATE,$data->TIME),
        value => [$data->DX, $data->DY, $data->DZ],
        method => $data->METH,
        accfactor => $data->EFAC,
        filename => $self->file->name,
        lineno => $csv->lineno,
        recordno => $csv->recordno,
        comment => $data->COMM
        );
    $self->context->add('vecc_obs',new List()) if ! $self->context->has('vecc_obs');
    $self->context->vecc_obs->add($obs);
}


package ContractMark;
use base 'AutoHash';

package MarkReliability;
use base 'AutoHash';

package MdfcFileCheck;
use base 'CsvFileCheck';

sub new
{
    my ( $class, $context ) = @_;
    my $self =  bless CsvFileCheck->new($context,'MDFC1'), $class;
    $self->{coordsys} = $self->context->cslist->coordsys('nzgd2000');
    return $self;
}


sub checkrecord
{
    my($self,$data) = @_;
    my $csv = $self->csv;
    my $mark = new ContractMark (
        code => $data->CODE,
        # status => $data->MRKS,
        type => $data->MRKT,
        exists => uc($data->EXMK) eq 'Y',
        district => $data->DISTRICT,
        order => $data->CROD,
        crd => $self->{coordsys}->coord($data->ORDV1,$data->ORDV2,$data->ORDV3),
        name => $data->NAME,
        altname => $data->ALTN,
        comment => $data->COMM,
        filename => $self->file->name,
        lineno => $csv->lineno,
        recordno => $csv->recordno,
        );
    $self->context->add('mdfc_marks',new MarkList()) if ! $self->context->has('mdfc_marks');
    if( $self->context->mdfc_marks->mark($data->CODE))
    {
        $self->report_error('Mark data for ',$data->CODE,' duplicated');
    }
    else
    {
        $self->context->mdfc_marks->add($mark);
        $self->context->marks->markhash($data->CODE)->add('mdfc',$mark);
    }
}

package RlbcCheckObs;
use base 'AutoHash';

package RlbcCheckData;
use base 'AutoHash';

package RlbcFileCheck;
use base 'CsvFileCheck';

sub _convertAngle
{
    my($angle)=@_;
    return undef if $angle !~ /(\-)?(\d+)\.([0-5]\d)([0-5]\d)$/;
    return ($1 eq '-' ? -1 : 1) * ($2 + $3/60.0 + $4/3600.0);
}

sub new
{
    my ( $class, $context ) = @_;
    my $self =  bless CsvFileCheck->new($context,'RLBC1'), $class;
    $context->add('reliability_data',{}) if ! $context->has('reliability_data');
    return $self;
}

sub checkrecord
{
    my($self,$data) = @_;
    my $csv = $self->csv;
    my $code=$data->CODE;
    my $type=$data->TYPE;
    my $relbdata=$self->context->{reliability_data};
    $relbdata->{$code}= new RlbcCheckData( 
        code=>$code,
        obs=>[]
        )
        if ! exists $relbdata->{$code};
    my $obs=new RlbcCheckObs(
        type=>$type,
        fref=>$data->FREF,
        tref=>$data->TREF,
        bear=>_convertAngle($data->BEAR),
        dist=>$data->DIST,
        bsng=>_convertAngle($data->BSNG || '0.0000'),
        pref=>$data->PREF,
        csys=>$data->CSYS,
        date=>$data->DATE,
        stpe=>$data->STPE
        );
    push(@{$relbdata->{$code}->obs},$obs);
}

package MaintenanceMark;
use base 'AutoHash';

package MwrcFileCheck;
use base 'CsvFileCheck';

sub new
{
    my ( $class, $context ) = @_;
    my $self =  bless CsvFileCheck->new($context,'MWRC2'), $class;
    return $self;
}

sub checkrecord
{
    my($self,$data) = @_;
    my $maintained = 0;
    foreach my $field (qw/MPSM MPSB MPSP/)
    {
        next if $field eq 'MPSB' && $data->MRKE eq 'NB';
        my $maint = lc($data->value($field));
        if( $maint !~ /^\s*(none|no\s+maintenance.*)?\s*$/ )
        {
            $maintained = 1;
            last;
        }
    }
    my $mark = new MaintenanceMark (
        code => $data->CODE,
        data => $data,
        maintained => $maintained,
        filename => $self->file->name,
        lineno => $self->csv->lineno,
        recordno => $self->csv->recordno,
    );
    $self->context->add('mwrc_marks',new MarkList()) if ! $self->context->has('mwrc_marks');
    if( $self->context->mwrc_marks->mark($data->CODE))
    {
        $self->report_error('Mark data for ',$data->CODE,' duplicated');
    }
    else
    {
        $self->context->mwrc_marks->add($mark);
        $self->context->marks->markhash($data->CODE)->add('mwrc',$mark);
    }
}


package SnapFileCheck;
use base 'FileCheck';
use SNAP::CrdFile;
use SNAP::DataFile;

sub new
{
    my($class,$context) = @_;
    return bless FileCheck->new($context), $class;
}


sub checkfile
{
    my($self,$file) = @_;
    my $name = $file->name;
    if( $name =~ /\.dat$/i ){ $self->check_datafile($file); }
    if( $name =~ /\.crd$/i ){ $self->check_crdfile($file); }
}

sub check_crdfile
{
    my($self,$file) = @_;
    eval
    {
        my $crds = SNAP::CrdFile->open($file->path);
        $self->context->add('snapcrds',{}) if ! $self->context->has('snapcrds');
        $self->context->snapcrds->{$file->name} = $crds;
    };
    if( $@ )
    {
        $self->report_error("Error loading SNAP coordinate file: ",$@);
    }
}

sub check_datafile
{
    my($self,$file) = @_;
    eval
    {
        $self->context->add('snap_obs',new List()) if ! $self->context->has('snap_obs');
        my $snap_obs = $self->context->snap_obs;

        my $of = SNAP::DataFile->open($file->path);
        while( my $obsset = $of->next )
        {
            foreach my $o (@{$obsset->obs})
            {
                $snap_obs->add($o) if $o->type eq 'GPS';
            }
        }
    };
    if( $@ )
    {
        $self->report_error("Error loading SNAP observation file: ",$@);
    }
}

1;

