# Module to manage the list of contract files

use strict;
use File::Find;
use File::Spec;
use ContractUtil;
use ContractConfig;

package ContractFile;

sub new
{
    my($class,$path) = @_;
    my $name = $path;
    $name =~ s/.*[\\\/]//;
    return bless 
    {
        path=>$path,
        name=>$name,
        filetype=>undef,
        type=>'',
    }, $class;
}

sub setFileType 
{ 
    my($self,$filetype) = @_;
    $self->{filetype} = $filetype;
    $self->{type} = $filetype->type;
}

sub path { return $_[0]->{path};}
sub name { return $_[0]->{name};}
sub type { return $_[0]->{type};}
sub filetype { return $_[0]->{filetype};}

#######################################################

package ContractJobType;

our $_types = {};
our $_cfg = undef;
our $_cfgLoaded = 0;

sub _create
{
    my($type,$name,$reqfiles) = @_;
    my $self = bless { 
        type=>$type, 
        name=>$name, 
        reqfiles=>$reqfiles, 
    }, 
    'ContractJobType';
    $_types->{$type} = $self;
    return $self;
}

sub type { return $_[0]->{type}; }
sub name { return $_[0]->{name}; }
sub reqfiles { return $_[0]->{reqfiles}; }

sub loadConfig
{
    return if $_cfgLoaded;
    my($cfg) = @_;
    $_cfgLoaded = 1;
    $cfg = new ContractConfig() if ! $cfg;
    $_cfg = $cfg;
    foreach my $key ($cfg->keys)
    {
        next if $key !~ /^([A-Z0-9]+)_JOB_TYPE_NAME$/i;
        my $type = uc($1);
        my $name = $cfg->get($type.'_JOB_TYPE_NAME',$type);
        my @reqfiles = split(' ',$cfg->get($type.'_JOB_REQ_FILES',''));
        _create($type,$name,\@reqfiles)
    }
}

sub jobTypeFromName
{
    my ($name)=(@_);
    loadConfig();
    my $namelen = length($name);
    my $results = undef;
    foreach my $t (values %$_types)
    {
        if(uc(substr($t->name ,0,$namelen)) eq uc($name))
        {
            return undef if $results;
            $results = $t;
        }
    }
    return $results;
}

#######################################################


package ContractFileType;

our $_types = {};
our $_cfg = undef;
our $_cfgLoaded = 0;
our $_nameCodesRe = qr/./;
our $_nameCodeRegex = {};

sub _create
{
    my($type,$name,$re,$cols,$cond,$defined) = @_;
    $name ||= $type;
    my $typere = _filenameRegex($re);
    my $hasre = $re ne '' ? 1 : 0;
    my $self = bless { 
        type=>$type, 
        name=>$name, 
        re=>$typere, 
        cols=>[],
        csvopts=>{},
        tests=>[],
        hasre=>$hasre,
        defined=>$defined }, 
        'ContractFileType';

    $self->readColumns( $cols ) if $cols;
    $self->readTests( $cond ) if $cond;
    $_types->{$type} = $self;
    return $self;
}

sub loadConfig
{
    return if $_cfgLoaded;
    my($cfg) = @_;
    $_cfgLoaded = 1;
    $cfg = new ContractConfig() if ! $cfg;
    loadFilenameRegex($cfg->{file_name_regexp});
    $_cfg = $cfg;
    foreach my $key ($cfg->keys)
    {
        next if $key !~ /^([A-Za-z0-9]+)_TYPE_NAME$/i;
        my $type = uc($1);
        my $name = $cfg->get($type.'_TYPE_NAME',$type);
        my $re = $cfg->get($type.'_FILE_NAME','');
        my $cols = $cfg->get($type.'_FILE_COLS','');
        my $cond = $cfg->get($type.'_CONDITIONAL_TESTS','');
        _create($type,$name,$re,$cols,$cond,1);
    }
}

sub loadFilenameRegex
{
    my($regexlist)=@_;
    foreach my $regex (split(/\n/,$regexlist))
    {
        next if $regex =~ /^\s*$/;
        if( $regex =~ /^\s*(\w+)\s+(\S+)\s*$/ )
        {
            $_nameCodeRegex->{$1}=$2;
            next;
        }
        print("Invalid file name regex \"$regex\" in configuration\n");
    }
    my $re='('.join('|',keys %$_nameCodeRegex).')';
    $_nameCodesRe=qr/$re/;
}

sub inferredTypes
{
    my($name) = @_;
    loadConfig();
    my @types = ();
    foreach my $type (values %$_types)
    {
        next if ! $type->hasre;
        next if ! $type->nameIsValid($name);
        push(@types,$type) if $type->type !~ /^MIS/;
    }
    return \@types;
}

sub inferredType
{
    my($name) = @_;
    my $types = inferredTypes($name);
    # If only one valid type return that
    return $types->[0] if scalar(@$types) == 1;
    # If one non-misc type, return that
    my @spectypes = grep {$_->{type} !~ /^MISC/} @$types;
    # Otherwise cannot be inferred
    return undef;
}

sub filetype 
{
    my($type) = @_;
    loadConfig();
    $type = uc($type);
    return $_types->{$type} if exists $_types->{$type};
    my $basetype = $type;
    return $_types->{$basetype} if $basetype =~ s/\d$// && exists $_types->{$basetype};
    return _create($type);
}

sub type { return $_[0]->{type}; }
sub typeid { my $t = $_[0]->{type}; return $t =~ /\d$/ ? $t : $t.'1'; }
sub name { return $_[0]->{name}; }
sub hasre { return $_[0]->{hasre}; }
sub defined { return $_[0]->{'defined'}; }
sub csvopts { return $_[0]->{'csvopts'}; }

sub _filenameRegex
{
    my($pattern) = @_;
    my $re = '';
    if( $pattern )
    {
        my $join = '^(';
        foreach my $name (split(' ',$pattern))
        {
            $name =~ s/\./\\./g;
            $name =~ s/\*/.*/g;
            $name =~ s/$_nameCodesRe/$_nameCodeRegex->{$1} || $1/eg;
            $re .= $join;
            $re .= $name;
            $join = '|';
        }
        $re .= ')$' if $re;
    }
    else
    {
        $re = '.?';
    }
    return qr/$re/i;
}


sub createTest
{
    my( $self, $name, $regex ) = @_;
    my $message = "Invalid value \"%%\" for $name field";
    $regex =~ s/\s+$//;
    my $subtest = undef;
    if( $regex =~ /^codes\s*(\s*\w+\s*)(\?)?$/i )
    {
        $regex = $_cfg->codeRegexp($1,$2);
        $message = "$name value \"%%\" is not a valid $1 code";
    }
    elsif( $regex =~ /^(number|integer)(?:\s*\((\-?\d+\.?\d*)?\s*\,\s*(\-?\d+\.?\d*)?\))?(\?)?$/i )
    {
        my ($type,$vmin, $vmax, $optional ) = (lc($1),$2,$3,$4);
        $regex = "";
        $regex = "\\.?\\d*" if $type eq 'number';
        $regex = "^(?:\\-?\\d+$regex)$optional\$";
        $message = "$name field \"%%\" is not " . ($type eq 'number' ? 'a number' : 'an integer');
        my @tests = ();
        push(@tests,"\$_[0] < \$vmin") if $vmin ne '';
        push(@tests,"\$_[0] > \$vmax") if $vmax ne '';
        if(@tests)
        {
            my $test = join(" || ",@tests);
            my $rangemsg = "$name value \"%%\" is outside the valid range ($vmin - $vmax)";
            my $sub = "sub { return \$rangemsg if \$_[0] ne '' && ($test)}";
            $subtest = eval $sub;
        }
    }
    elsif( $regex =~ /^gcode(\?)?$/i )
    {
        $regex = "^([0-9A-Z]{4})$1\$";
        $message = "$name value \"%%\" is not a valid geodetic code";
    }
    elsif( $regex =~ /^(?:gps)?date(\?)?$/i )
    {
        $regex = "^((18|19|20)[0-9]{2}\.(?:(0[1-9]|1[0-2])\.(3[01]|[12][0-9]|0[1-9])|([0-9]{3})))$1\$";
        $message = "$name value \"%%\" is not a valid date";
    }
    # Now allowing all dates to be either GPS day of year format or single date
    # elsif( $regex =~ /^date(\?)?$/i )
    # {
    #     $regex = "^((18|19|20)[0-9]{2}\.(0[1-9]|1[0-2])\.(3[01]|[12][0-9]|0[1-9]))$1\$";
    #     $message = "$name value \"%%\" is not a valid date";
    #  }
    elsif( $regex =~ /^time(\?)?$/i )
    {
        $regex = "^([0-2]?[0-9](?:[.:](?:[0-5][0-9])?)?)$1\$";
        $message = "$name value \"%%\" is not a valid time";
    }
    my $re = qr//;
    eval 
    {
        $re = qr/$regex/;
    };
    if( $@ )
    {
        print "Invalid regular expression $regex in definition of $name field\n";
    }

    my $message2 = "$name value is missing";
    return sub 
    {
        my $v=$_[0]->value($name);
        my $result='';
        if($v !~ $re){ $result= ($v eq '') ? $message2 : $message; }
        if( $subtest && ! $result )
        {
            $result = $subtest->($v);
        }
        $result =~ s/\%\%/$v/;
        return $result;
    }
}


sub readColumns
{
    my($self,$cols) = @_;
    my @tests = ();
    if( $cols )
    {
        foreach my $col (split(/\n/,$cols) )
        {
            next if $col =~ /^\s*\#/;
            next if $col =~ /^\s*$/;
            my($name,$regex) = split(' ',$col,2);
            next if ! $name;
            if( $name =~ /(.*?)\?(?:=(.*))?/ )
            {
                $name=$1;
                $self->{csvopts}->{default_cols}->{$name}=$2;
            }
            push ( @{$self->{cols}}, $name );

            if ( $regex )
            {
                my $testsub = $self->createTest( $name, $regex );
                my $test = { cols=>[$name], test=>$testsub };
                push(@{$self->{tests}}, $test );
            }
        }
    }
}



# Load tests for a data set.  These are defined by records with five fields
#    cond_col    Column used to determine if test applieds
#    cond_re     Test regular expression - test applies if con_col matches
#                Can be preceded with ! to negate
#    test_col    Column to be validated
#    test_re     Regular expression that must be matched
# Creates a sub that takes a record definition - assumed to have a value()
# function, and returns an error message

sub readTests
{
    my($self,$cond) = @_;
    my @tests = ();
    if( $cond )
    {
        foreach my $cnd (split(/\n/,$cond) )
        {
            next if $cnd =~ /^\s*\#/;
            next if $cnd =~ /^\s*$/;

            my($cfield,$cre,$tfield,$tre) = split(' ',$cnd,4);
            next if ! $tre;

            my $cmatch = '=~';
            if( $cre =~ /^\!/){ $cmatch = '!~'; $cre = substr($cre,1); }
            my $re = qr//;
            eval
            {
                $re = qr/$cre/;
            };
            if( $@ )
            {
                print "Invalid conditional expression $cre for field $cfield\n";
            }
            my $sub = "sub { return \$_[0]->value(\$cfield) $cmatch \$re; }";
            my $condsub = eval $sub;
            my $testsub = $self->createTest($tfield,$tre );
            my $test = { cols=>[$tfield,$cfield], test=>$testsub, condition=>$condsub };
            push(@{$self->{tests}}, $test );
        }
    }
    return \@tests
}


sub nameIsValid
{
    my($self,$name) = @_;
    return $name =~ $self->{re};
}

sub columnNames
{
    my($self) = @_;
    return wantarray ? @{$self->{cols}} : $self->{cols};
}


# Returns an error and a hash keyed by missing column names

sub missingCsvColumns
{
    my($self,$csv) = @_;
    my @missing = ();
    my %missingcols = ();
    foreach my $name (@{$self->{cols}})
    {
        next if defined($csv->colno($name));
        $missingcols{uc($name)} = 1;
        push(@missing,$name);
    }
    return undef if ! @missing;
    my $error = "Missing column";
    $error .= "s" if @missing > 1;
    $error = join(" ",$error,@missing);
    
    return $error,\%missingcols;
}

# Assumes $record has function value(colname) to retrieve column

sub errorsInRecord
{
    my($self,$record,$missingcols) = @_;
    my @errors = ();
    my %failed = ();
    foreach my $test (@{$self->{tests}})
    {
        my $usetest = 1;
        foreach my $col (@{$test->{cols}})
        {
            $usetest = 0 if $missingcols->{$col} || $failed{$col};
        }
        next if ! $usetest;
        next if $test->{condition} && ! $test->{condition}->($record);
        my $error = $test->{test}->($record);
        next if ! $error;
        push(@errors,$error);
        $failed{$test->{cols}->[0]} = 1;
    }
    return @errors if wantarray;
    return \@errors if @errors;
    return undef;
}

sub set_colno { $_[0]->{colno} = $_[1]; }
sub colno { return $_[0]->{colno} };


#######################################################

package ContractFileList;

sub new
{
    my($class,$base,$cfg,$indexname,$overwrite_index) = @_;
    $base ||= '.';

    ContractFileType::loadConfig($cfg);
    my $files = [];
    my $fileindex = {};
    my @indexfiles = ();
    my $indextype = ContractFileType::filetype('idx');
    my @errors = ();

    $indexname = lc($indexname);
    my $indexfile = undef;
    my $lcnames={};
    
    File::Find::find( sub
        {
            return if ! -f $_;
            my $fullpath=$File::Find::name;
            my $relpath=File::Spec->abs2rel($fullpath,$base);
            my $file = new ContractFile($fullpath);
            push(@$files,$file);
            my $name = lc($file->name);
            push( @errors, "$name is duplicated")
                if exists $lcnames->{$name};
            $lcnames->{$name}=$file;
            $fileindex->{$relpath} = $file;
            if( $name eq $indexname || (! $indexname && $indextype->nameIsValid($name)))
            {
                push(@indexfiles,$file) if ! $overwrite_index;
            }
        }, $base);

    if( ! @indexfiles && ! $overwrite_index )
    {
        my @columns = $indextype->columnNames;
        foreach my $file (@$files)
        {
            next if $file->name !~ /\.csv$/i;
            eval
            {
                my $csv = new CSV_File($file->path);
                push(@indexfiles,$file)  if $csv->hascolumns(@columns);
                $csv->close();
            };
        }
    }

    my $self= bless
    {
        base=>$base,
        files=>$files,
        fileindex=>$fileindex,
        errors=>\@errors,
    }, $class;
    
    if( ! @indexfiles && $indexname && 
        (my $indexfile = $self->createIndexFile($indexname))) {
        $indexfile = new ContractFile($indexfile);
        push(@$files,$indexfile);
        push(@indexfiles,$indexfile);
        $fileindex->{$indexname} = $indexfile;
    }

    push( @errors, "Cannot find a contract index file in folder $base")
        if ! @indexfiles;

    push(@errors,"Not a unique index file: ".
        join(' ',map {$_->name} @indexfiles))
        if @indexfiles > 1;


    $self->{indexfile} = @indexfiles == 1 ? $indexfiles[0] : undef;
    return $self;
}

sub inferFileTypes
{
    my ($self) = @_;
    foreach my $f (@{$self->{files}})
    {
        next if $f->{type};
        my $type = ContractFileType::inferredType( $f->name() );
        next if ! $type;
    }

}

sub createIndexFile
{
    my($self,$indexname) = @_;
    $indexname = 'index.csv' if ! $indexname;
    $indexname .= '.csv' if $indexname !~ /\.csv$/i;
    my $basepath=$self->{base};
    my $indexfile = $basepath.'/'.$indexname;
    print "Creating index file ",$indexfile,"\n";
    open( my $fh, ">", $indexfile );
    if( ! $fh )
    {
        die "Cannot create index file $indexfile\n";
    }
    print $fh '"DAFL","DATE","SCNO","CLID","DAFT","COMM"',"\n";
    my @files= @{$self->{files}};
    my %filetypes=();
    foreach my $f (@files)
    {
        my $type = ContractFileType::inferredType( $f->name() );
        next if ! $type;
        next if $type->type eq 'IDX';
        next if $type->type eq 'IGNORE';
        $filetypes{$f->name()}=$type->typeid();
    }
    @files = sort { ($filetypes{$a->name} cmp $filetypes{$b->name}) || $a->name cmp $b->name } @files;

    foreach my $f (@files)
    {
        next if ! $filetypes{$f->name};
        my ($d,$m,$y) = (localtime((stat($f->path()))[9]))[3,4,5];
        my $date = sprintf("%04d.%02d.%02d",$y+1900,$m+1,$d);
        my $scno = sprintf("%04d999",$y+1900);
        my $fpath=File::Spec->abs2rel($f->path,$basepath);
        print $fh join(",", map { '"'.$_.'"' } (
            $fpath,$date,$scno,'',$filetypes{$f->name},'')),"\n";
    }
    close($fh);

    return $indexfile;
}

sub file
{
    my($self,$name) = @_;
    return $self->{fileindex}->{$name};
}

sub find 
{
    my($self,$namere,$type) = @_;
    $namere = '.' if ! $namere;
    eval
    {
        $namere = qr/$namere/i;
    };
    return undef if $@;
    my @found = ();
    foreach my $file ($self->files)
    {
        next if $type && $type ne $file->type;
        next if $file->name !~ $namere;
        push(@found,$file);
    }
    return @found if wantarray;
    return \@found if @found;
    return undef;
}

sub files { return wantarray ? @{$_[0]->{files}} : $_[0]->{files}; }
sub indexfile { return $_[0]->{indexfile}; }
sub errors { return wantarray ? @{$_[0]->{errors}} : $_[0]->{errors}; }

1;
