
use strict;
use Data::Dumper;

package ContractUtil;
use POSIX;

# Convert contract format dates/times to timestamp

sub CalcDate
{
    my($date,$time) = @_;
    if( $date =~ /^(\d+)\.(\d+)$/)
    {
        $date = mktime(0,0,0,$2,0,$1-1900);
    }
    elsif( $date =~ /^(\d+)\.(\d+)\.(\d+)$/ )
    {
        $date=mktime(0,0,0,$3,$2-1,$1-1900);
    }
    else
    {
        return 0;
    }
    if( $time=~/^(\d+)[\:\.](\d+)(?:[\:\.](\d+))?$/ )
    {
        $date += ($1*60+$2)*60+$3;
    }
    return $date;
}

# Convert contract format dates to YMD

sub YMD_Date
{
    my ($date) = @_;
    return $date if $date !~ /^(\d+)\.(\d+)$/;
    my $dt = mktime(0,0,0,$2,0,$1-1900);
    my ($y,$m,$d) = (localtime($dt))[5,4,3];
    $m++;
    $y+=1900;
    return sprintf("%04d.%02d.%02d",$y,$m,$d);
}

package List;

# Wrapped array hash with add function

sub new
{
    my($class) = @_;
    return bless [],$class;
}

sub add
{
    my($self,$value) = @_;
    push(@{$self},$value);
}


package AutoHash;

# Hash array class with autogenerated ->member functions.  
# 

our $AUTOLOAD;
use Scalar::Util 'reftype';

sub new
{
    my($class) = shift(@_);
    my %self = (@_);
    $self{_missing}='print' if ! exists $self{_missing};
    $self{_createmember}=1 if ! exists $self{_createmember};
    return bless \%self, $class;
}

sub AUTOLOAD
{
    my( $self, $value ) = @_;
    my $key = $AUTOLOAD;
    my $package = $AUTOLOAD;
    $key =~ s/.*://;   # strip fully-qualified portion
    $package =~ s/::[^:]*$//;   # strip name
    return if $key eq 'DESTROY';

    die "Invalid AUTOLOAD of $key on $package - \$self not a hash\n"
       if reftype($self) ne 'HASH';
    $self->{$key} = $value if $key =~ s/^set_//;
    if(! exists $self->{$key} )
    {
        my $error = "Invalid method call $key in $package: caller ".join(' ',caller(0))."\n";
        die $error if $self->{_missing} eq 'die';
        print $error if $self->{_missing} eq 'print';
    }
    eval "sub $key { return \$_[0]->{$key} }" if $self->{_createmember};
    return $self->{$key};
}


sub add
{
    my($self,$key,$value) = @_;
    $self->{$key} = $value;
}

sub has
{
    my($self,$key) = @_;
    return exists $self->{$key};
}

sub get
{
    my($self,$key) = @_;
    return $self->{$key};
}

package MarkHash;
use base 'AutoHash';

sub new
{
    my($self,$code) = @_;
    return new AutoHash code=>$code;
}


package MarkList;

# Mark list - list of hash ref objects with ->{code} element. 
# Maintains an index of codes. 
# Returns list sorted by code

sub new
{
    my ($class) = @_;
    return bless { marks=>[], code=>{}, duplicate=>[], sorted=>1 }, $class;
}

sub add
{
    my($self,$mark) = @_;
    my $code = uc($mark->{code} || $mark->{CODE});
    die "Invalid mark added to Marklist\n".Dumper($mark) if ! $code;
    if(exists $self->{code}->{$code})
    {
        push(@{$self->{duplicate}},$mark);
    }
    else
    {
        push(@{$self->{marks}},$mark);
        $self->{code}->{$code} = $mark;
    }
    $self->{sorted} = 0;
}

sub codes
{
    my($self) = @_;
    my @codes = sort keys %{$self->{code}};
    return wantarray ? @codes : \@codes;
}

sub marks 
{ 
    my($self) = @_;

    if( ! $self->{sorted} )
    {
        my @marks = map { $self->mark($_) } @{$self->codes()};
        $self->{marks} = \@marks;
        $self->{sorted} = 1;
    }
    return wantarray ? @{$self->{marks}} : $self->{marks}; 
}

sub mark { return $_[0]->{code}->{uc($_[1])}; }

sub markhash
{
    my( $self, $code ) = @_;
    my $mark = $self->mark($code);
    return $mark if $mark;
    $mark = new MarkHash($code);
    $self->add($mark);
    return $mark;
}

package Context;
use base 'AutoHash';

sub new
{
    my( $cls,$cfg,$files,$jobtype) = @_;
    return bless new AutoHash
    ( 
        cfg=>$cfg, 
        files=>$files,
        jobtype=>$jobtype,
        marks=>new MarkList(),
        missinginfo=>[],
        verbose=>1,
        debug=>0
    ), $cls;
}

sub record_mark_error
{
    my($self,$code,@errors) = @_;
    my $type = 'error';
    if( $errors[0] eq 'info'){ $type = shift(@errors); }
    my $mark = $self->marks->markhash($code);
    $mark->add('errors',[]) if ! $mark->has('errors');
    push(@{$mark->errors},new AutoHash(type=>$type,error=>join('',@errors)));
}

sub add_mark_image
{
    my($self,$code,$imagetype,$imagefile) = @_;
    my $mark = $self->marks->markhash($code);
    $mark->add('images',[]) if ! $mark->has('images');
    push(@{$mark->images},new AutoHash(type=>$imagetype, file=>$imagefile));
    
}


1;
