use strict;


package ReportFile;
use HTML::Entities;

our $files;
$files = {};

sub new
{
    my($class,$title,$name,$scripts) = @_;

    my $filename = $name || 'report';
    $filename =~ s/\.html$//i;

    if( exists $files->{$filename} )
    {
        $files->{$filename}++;
        $filename .= sprintf("_%03d",$files->{$filename});
    }
    else
    {
        $files->{$filename} = 0;
    }

    $filename .= ".html";
    open(my $fh,">$filename") || die "Cannot open report file $filename\n";
    $title ||= 'Contract dataset report';

    my $self= bless
    {
        filename => $filename,
        title => $title,
        scripts => $scripts,
        fh => $fh,
        div => 0,
        block => '',
        errorcount => 0,
        codeurl=>'{code}',
    }, $class;

    my $etitle = $self->escape($title);

    $self->_print(<<"EOD");
<html>
  <head>
    <title>$etitle</title>
    $scripts
  </head>
  <body>
EOD
    return $self;
}

sub setCodeUrl 
{ 
    my ($self,$codeurl) = @_;
    if( $codeurl =~ /\{code\}/ )
    {
        $codeurl = "<a class=\"codelink\" href=\"$codeurl\">{code}</a>";
    }
    else
    {
        $codeurl = '{code}';
    }
    $self->{codeurl} = $codeurl;
}

sub filename { return $_[0]->{filename}; }

sub close
{
    my($self) = @_;
    return if ! $self->{fh};
    $self->_endblock;
    $self->enddiv while $self->{div};
    $self->_print(<<"EOD");
  </body>
</html>
EOD
    close($self->{fh});
    delete $self->{fh};
}

sub DESTROY
{
    my($self) = @_;
    $self->close();
}

sub link
{
    my($self,$text) = @_;
    $text ||= $self->{filename};
    return "<a href=\"$self->{filename}\">$text</a>";
}

sub escape
{
    my($self,$text) = @_;
    return encode_entities($text);
}

sub write_title
{
    my($self,$title,$level) = @_;
    $title ||= $self->{title};
    $level ||= 1;
    my $block = "h$level";
    $self->_startblock($block);
    $self->write($title);
    $self->_endblock();
}

sub table
{
    my($self,$class) = @_;
    my $block = 'table';
    $block .= " class=\"$class\"" if $class;
    $self->_startblock($block);
}

sub paragraph
{
    my($self,$class) = @_;
    my $block = 'p';
    $block .= " class=\"$class\"" if $class;
    $self->_startblock($block);
}

sub write
{
    my($self) = shift(@_);
    $self->write_html(map {$self->escape($_)} @_);
}

sub write_html
{
    my($self) = shift(@_);
    $self->_startblock('p',1) if $self->{autopara};
    $self->_print(@_);
    $self->_endblock if $self->{autopara};
}

sub write_literal
{
    my($self) = shift(@_);
    $self->_startblock('pre');
    $self->_print(map {$self->escape($_)} @_);
}

sub write_table
{
    my($self) = shift(@_);
    $self->write_table_html( map {$self->escape($_)} @_ );
}

sub write_table_html
{
    my($self) = shift(@_);
    $self->_startblock('table');
    $self->_print("<tr>\n");
    foreach my $col (@_) { $self->_print("<td>$col</td>\n"); }
    $self->_print("</tr>\n");
}

sub autoparagraph
{
    my($self,$option) = @_;
    $option = 1 if ! defined($option);
    $self->{autopara} = $option;
}

sub div
{
    my($self,$class) = @_;
    $class = " class=\"$class\"" if $class;
    $self->_print("<div$class>\n");
    $self->{div}++;
}

sub enddiv
{
    my($self) = @_;
    $self->_endblock;
    $self->_print("</div>\n") if $self->{div};
    $self->{div}--;
}

sub _startblock
{
    my($self,$type,$check)=@_;
    my $tag = $type;
    $tag =~ s/\s.*//;
    return if $check && $self->{block} eq $tag;
    $self->_endblock;
    $self->{block} = $tag;
    $self->_print("<$type>");
}

sub _codeurl
{
    my($self,$code) = @_;
    my $codeurl = $self->{codeurl};
    $codeurl =~ s/\{code\}/$code/xg;
    return $codeurl;
}

sub _print
{
    my $self = shift @_;
    my $fh = $self->{fh};
    return if ! $fh;
    my $line = join('',@_);
    $line =~ s/\{code:(\w\w\w\w)\}/$self->_codeurl($1)/eg;
    print $fh $line;
}

sub _endblock
{
    my($self) = @_;
    return if ! $self->{block};
    $self->_print("</",$self->{block},">\n");
    $self->{block} = '';
}

sub section
{
    my($self,$title) = @_;
    $self->paragraph("section_header");
    $self->write($title);
    $self->_endblock;
}

sub start_errors
{
    my($self,$section,$description) = @_;
    $self->section($section);
    $self->{errorcount} = 0;
    $self->div("error");
    if( $description )
    {
        my @parts = split(/\n\n/,$description);
        foreach my $part (@parts )
        {
            $self->paragraph("section_description");
            $self->write( $part );
            $self->_endblock;
        }
    }
    $self->{counting} = 1;
}

sub end_errors
{
    my($self) = @_;
    $self->enddiv;
    $self->{counting} = 0;
    return $self->{errorcount};
}

sub error_count
{
    return $_[0]->{errorcount} || 0;
}

sub start_summary
{
    my($self,$section) = @_;
    $self->div("summary");
}

sub end_summary
{
    my($self) = @_;
    $self->enddiv;
}

sub write_info
{
    my $self = shift(@_);
    $self->paragraph("info");
    $self->write(@_);
}

sub write_error
{
    my $self = shift(@_);
    $self->paragraph("error");
    $self->write(@_);
    $self->{errorcount}++ if $self->{counting};
}

sub write_error_detail
{
    my $self = shift(@_);
    $self->paragraph("error detail");
    $self->write(@_);
}

sub write_info_detail
{
    my $self = shift(@_);
    $self->paragraph("info detail");
    $self->write(@_);
}

1;
