# Script to process a set of configuration files, looking for differences
# from contract specifications.
#
# Results are generated into a set of HTML report files.

use strict;

use FindBin;
use POSIX;
use lib $FindBin::Bin;
use lib $FindBin::Bin.'/lib';
use ReportFile;
use File::Find;
use File::Spec;
use LocalCoordSysDef;
use LINZ::Geodetic::CoordSysList;
use ContractConfig;
use ContractUtil;
use GdbMarkDetail;
use ContractFiles;
use ContractFileChecks;
use ContractChecks;
use HTML::Template;

use Getopt::Std;

my %opts;
getopts('x:r:i:j:nomIO',\%opts);

my $cfgext = $opts{c};
my $rptfile = $opts{r} || 'report';
$rptfile =~ s/\.html$//i;
my $idxfile = $opts{i};
$idxfile='index.csv' if ($opts{I} || $opts{O}) && ! $idxfile;
my $connect = $opts{n} ? 0 : 1;
my $mark_reports = $opts{m};
my $overwrite_index = $opts{o} | $opts{O};
my $jobtypename = $opts{j} || 'survey';


my $syntax = <<EOS;

con_check.pl: Checks contract data sets

syntax: con_check [options] contract_data_directory

options are:
  -j job_type       Job type, one of survey (the default), maintenance, height, 
                    inventory.
  -r report_file    Specifies the name of the report file (default report.html).
                    Only specify the name (.html will be added)
  -i index_file     Specify the name of the index file.  Selects the file to
                    use if it is ambiguous or not the default file.
                    Also creates the index file if it does not exist.
  -I                Shorthand for -i index.csv
  -o                Overwrite the index file
  -O                Shorthand for -i index.csv -o
  -n                Don't connect to the geodetic database
  -m                Generate a report for each mark

EOS

die $syntax if ! @ARGV;

###############################################
#  Load configuration

my $cfg = new ContractConfig($cfgext);
my $jobtype = ContractJobType::jobTypeFromName($jobtypename)
  || die "Invalid job type $jobtypename specified\n";

print "Job type: ",$jobtype->name,"\n";

my $cslist = LINZ::Geodetic::CoordSysList->newFromCoordSysDef();

my $gdb = new GdbMarkDetailList($cfg, connect=>$connect);

my $basedir = $ARGV[0] || '.';
$basedir = File::Spec->rel2abs($basedir);

# Load the files found under the base directory
# and identify the index file.
# Set up the context for the checking.

my $contractfiles = new ContractFileList($basedir,$cfg,$idxfile,$overwrite_index);

my $context = new Context($cfg,$contractfiles,$jobtype);
$context->add('cslist',$cslist);

# Create the top level report file.  Get the scripts to be included in
# the report header.

my $htmlscriptfile = $cfg->get('report_scripts_file',$0.".htmlscript");
my $htmlcssfile = $cfg->get('report_css_file',$0.".htmlcss");
my $marktemplatefile = $cfg->get('mark_template_file',$0.".marktemplate");
my $gdburl = $cfg->get('gdburl',"https://www.geodesy.linz.govt.nz/gdb/");
my $gdbimageurl = $cfg->get('gdbimageurl',"https://www.geodesy.linz.govt.nz/gdb/gdbimage.cgi");
my $mark_base = 'mark_';
my $mark_index = $mark_base.'index.html';
$mark_reports = 0 if ! -r $marktemplatefile;

my $script = get_file($htmlscriptfile); 
my $css = get_file($htmlcssfile);

my $report = new ReportFile('Geodetic contract report',$rptfile,$css.$script);
$context->set_toplevelreport( $report );
if( $mark_reports )
{
    $report->setCodeUrl($mark_base.'{code}.html');

}
else
{
    $report->setCodeUrl($gdburl.'?code={code}');
}

if($mark_reports)
{
    $report->write_html("<div class=\"menubar\"><span class=\"button\"><a href=\"$mark_index\">Mark index</a></span></div>");
}
$report->write_title();

$report->autoparagraph;
$report->write('Contract type: ',$jobtype->name);
$report->write('Contract deliverable folder: ',$basedir);
$report->write('Report run at ',strftime("%H:%M:%S %d-%b-%Y",localtime) );

my @errors= $contractfiles->errors;
if( @errors )
{
    $report->start_errors("Errors in folder");
    foreach my $error (@errors)
    {
        $report->write_error($error);
    }
    $report->end_errors();
    $report->write_error($report->error_count," errors in folder");
}

# Run the checks

my $indexfile = $contractfiles->indexfile;

if( $indexfile )
{
    $report->write('Index file: ',$indexfile->name);
    IndexFileCheck->new($context)->check($indexfile,$report);
    MdfcFileCheck->new($context)->check_files('MDFC1',$report);
    MwrcFileCheck->new($context)->check_files('MWRC2',$report);
    VeccFileCheck->new($context)->check_files('VECC1',$report);
    RlbcFileCheck->new($context)->check_files('RLBC1',$report);
    # Snap file checks no longer required - using CSV files for data
    #SnapFileCheck->new($context)->check_files('SNPA1',$report);
}

my $connected=0;
eval
{
    print "Loading codes from the geodetic database\n";
    my @codes=$context->marks->codes;
    $gdb->loadcodes(\@codes);
    foreach my $code ($context->marks->codes)
    {
        my $markgdb = $gdb->markinfo($code);
        $context->marks->mark($code)->add('gdb',$markgdb);
    }
    $connected=1;
};
if($@)
{
        print "Cannot check mark data against geodetic database - $@\n";
}

ContractChecks::CheckMarkData($context,$report,$connected);
ContractChecks::CheckMaintenanceData($context,$report,$connected);
ContractChecks::CheckReliabilityData($context,$report,$connected);
ContractChecks::CheckDuplicateVectorObs($context,$report);
ContractChecks::CheckIndependentSessions($context,$report);
ContractChecks::CheckIndependentBaselines($context,$report);
# ContractChecks::CompareVectorObs($context,$report);

$report->start_errors("Summary of missing information");
my $nmissing = 0;
foreach my $missing( @{$context->missinginfo} )
{
    $report->write_error($missing);
    $nmissing++;
}
$report->end_errors();
$report->start_summary();
if( $nmissing )
{
    $report->write_error("$nmissing items missing\n");
}
else
{
    $report->write_info("All information was supplied\n");
}
$report->end_summary();

my $rpt_filename =$report->filename;
$report->close();

if( $mark_reports )
{
    my $report_name = $rpt_filename;
    my $report_base = $rpt_filename;
    $report_name =~ s/.*[\\\/]//;
    $report_base =~ s/[^\\\/]*$//;
    $report_base .= '/' if $report_base;
    my $codes =  $context->marks->codes;
    my $templatestr = get_file($marktemplatefile);
    $templatestr =~ s/\{([\w\.]+)\}/<TMPL_VAR ESCAPE=HTML NAME=$1>/g;
    $templatestr =~ s/\{url:([\w\.]+)\}/<TMPL_VAR ESCAPE=URL NAME=$1>/g;
    $templatestr =~ s/\{html:([\w\.]+)\}/<TMPL_VAR ESCAPE=NONE NAME=$1>/g;
    $templatestr =~ s/\{loop:([\w\.]+)\}/<TMPL_LOOP NAME=$1>/g;
    $templatestr =~ s/\{\/loop\}/<\/TMPL_LOOP>/g;
    $templatestr =~ s/\{if:([\w\.]+)\}/<TMPL_IF NAME=$1>/g;
    $templatestr =~ s/\{else:\}/<TMPL_ELSE>/g;
    $templatestr =~ s/\{\/if\}/<\/TMPL_IF>/g;
    my $template = HTML::Template->new( 
        scalarref=> \$templatestr,
        die_on_bad_params=>0,
        );

    my @errcodes = grep { $context->marks->mark($_)->has('errors') } @$codes;
    my %errlinks = ();
    my $i;
    foreach $i ( 0 .. $#errcodes )
    {
        $errlinks{$errcodes[$i]} = {
            prev=>($i > 0 ? $errcodes[$i-1] : '' ),
            next=>($i < $#errcodes ? $errcodes[$i+1] : '')
            };
    }
    foreach $i ( 0 .. $#$codes )
    {
        my $code = $codes->[$i];
        my $markfile = $report_base.$mark_base.$code.'.html';
        my $mark = $context->marks->mark($code);
        if( $mark->has('images'))
        {
            foreach my $image (@{$mark->images})
            {
                $image->add('url',&make_url($image->file,$markfile));
            }
        }
        my $markparams = &make_params($mark,'mark');
        $markparams->{code} = $code;
        $markparams->{nexturl} = $i == $#$codes ? '' : $mark_base . $codes->[$i+1].'.html';
        $markparams->{prevurl} = $i == 0 ? '' : $mark_base . $codes->[$i-1].'.html';
        $markparams->{reporturl} = $report_name;
        $markparams->{indexurl} = $mark_index;
        $markparams->{gdburl} = $gdburl;
        $markparams->{gdbimageurl} = $gdbimageurl;
        $markparams->{css} = $css;
        if( $errlinks{$code} )
        {
            my $links = $errlinks{$code};
            $markparams->{nexterrurl} = $links->{next} ? $mark_base.$links->{next}.'.html' : '';
            $markparams->{preverrurl} = $links->{prev} ? $mark_base.$links->{prev}.'.html' : '';
        }
        $template->clear_params();
        $template->param($markparams);

        open( my $fh,">",$markfile) || die;
        print $fh $template->output();
        close($fh);
    }

    my $ncol = ($cfg->get('concheck.markindex.cols',6)+0) || 6;
    my $index = new ReportFile('Geodetic contract mark index',$report_base.$mark_index,$css);
    $index->write_html("<div class=\"menubar\"><span class=\"button\"><a href=\"$report_name\">Summary</a></span></div>");
    $index->write_title();

    my $table = '';
    $table .= "<table class=\"mark_index\">\n";
    my $offset = int($#$codes/$ncol) + 1;
    foreach my $i (0..$offset-1)
    {
        $table .= "<tr>\n";
        my $nc = $i;
        foreach my $c (0..$ncol-1)
        {
            my $code = $codes->[$nc];
            $nc += $offset;
            $table .= "<td>";
            if( $code )
            {
                my $error = $context->marks->mark($code)->has('errors');
                $table .= '<span class="error">' if $error;
                $table .= "<a href=\"$mark_base$code.html\">$code</a>";
                $table .= '</span>' if $error;
            }
            $table .= "</td>";
        }
        $table .= "</tr>\n";
    }
    $table .= "</table>\n";
    $index->write_html($table);
    $index->close();
}

print "\nThe report is in ".$report->filename."\n\n";

# system("start \"".$report->filename."\"");


sub make_url
{
    my($file,$basefile) = @_;
    $basefile = File::Spec->rel2abs($basefile);
    $file = File::Spec->rel2abs($file);
    $basefile =~ s/[\\\/][^\\\/]*$//;
    my $url = File::Spec->abs2rel($file,$basefile);
    $url =~ s/\\/\//g;
    return $url;
}

sub make_params
{
    my($var,$prefix,$params) = @_;
    $params ||= {};

    my $type = ref($var);
    if( $type =~ /^CSV_File\:\:RecordType\d+$/ )
    {
        foreach my $f (@{$var->columns})
        {
            my $name = $prefix.($prefix ? '.' : '').$f;
            $params->{$name} = $var->value($f);
        }
    }
    elsif( $type eq 'HASH' || ($type =~ /[a-z]/ && $var->isa('HASH')) )
    {
        foreach my $k ( keys %$var )
        {
            next if $k !~ /^[a-zA-Z]/;
            my $name = $prefix.($prefix ? '.' : '').$k;
            &make_params( $var->{$k}, $name, $params );
        }
        $params->{$prefix} = 1 if $prefix;
    }
    elsif( $type eq 'ARRAY' || ($type =~ /[a-z]/ && $var->isa('ARRAY')) )
    {
        my $values = [];
        my $count = 0;
        foreach my $v (@$var)
        {
            push(@$values,&make_params($v,''));
            $count++;
        }
        my $name = $prefix || 'value';
        $params->{$name} = $values;
        $params->{$name.'_count'} = $count;
    }
    else
    {
        $var .= '';
        my $name = $prefix || 'value';
        $params->{$name} = $var;
    }
    return $params;
}
    

sub get_file
{
    my ($file) = @_;
    my $content = '';
    if( -r $file )
    {
        open( my $f, "<", $file );
        $content = join('',<$f>);
        close($f);
    }
    return $content;
}
